/*
   Instead of maintaining a List of Future results (that
   are being worked on by the threadpool), this example
   uses a CompletionService, which is a threadpool combined
   with an (internal) List of Future results.

   After submitting all the work to the CompletionService,
   we can retrieve from the CompletionService's internal
   List each completed result as soon as the result is ready.

   Another way to put this is, we retrieve from the
   CompletionService completed results in the order in
   which they are ready, not in the order in which we
   submitted them (as in the last example).
*/

import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.Future;
import java.util.concurrent.ExecutionException;

import java.util.List;

/**
   Execute 1,000 tasks by placing the tasks in a "task pool" that
   is worked on by a fixed number of threads from a "thread pool".
*/
public class ThreadPoolExample_saved_results_v3
{
   public static void main(String[] args)
   {
      final long startTime = System.currentTimeMillis();

      for (long i = 0; i < 1_000; ++i) // try 10_000
      {
         final long x = i + 50_000_000_000_000_001L;
         // submit a Callable to the CompletionService
         completionService.submit( ()->Primes.factorize(x) );
      }
      executor.shutdown();

      for (long i = 0; i < 1_000; ++i)
      {
         try
         {
            // retrieve a completed result from the CompletionService's internal List
            Future<List<Long>> result = completionService.take(); // this may block
            System.out.println( result.get() ); // this get() will not block!
         }
         catch(InterruptedException e){}
         catch(ExecutionException e){}
      }

      final long stopTime = System.currentTimeMillis();
      System.out.println("(" + (stopTime - startTime) + "ms)");
   }

   private static final int cores = Runtime.getRuntime().availableProcessors();
   private static final ExecutorService executor = Executors.newFixedThreadPool(cores);
   private static final CompletionService<List<Long>> completionService = new ExecutorCompletionService<>(executor);
}
